// react:
import { 
// hooks:
useState, } from 'react';
// cssfn:
import { 
// writes css in javascript:
rule, states, style, vars, imports, cssVars, } from '@cssfn/core'; // writes css in javascript
// reusable-ui utilities:
import { 
// hooks:
useEvent, } from '@reusable-ui/hooks'; // react helper hooks
import { useSemantic, } from '@reusable-ui/semantics'; // a semantic management system for react web components
import { 
// hooks:
usePropAccessibility, usePropActive, } from '@reusable-ui/accessibilities'; // an accessibility management system
// reusable-ui features:
import { 
// hooks:
usesAnimation, } from '@reusable-ui/animation'; // animation stuff of UI
// reusable-ui variants:
import { 
// hooks:
setOutlined, } from '@reusable-ui/outlineable'; // outlined (background-less) variant of UI
import { 
// hooks:
setMild, } from '@reusable-ui/mildable'; // mild (soft color) variant of UI
const [activatableVars] = cssVars();
{
    const { animationRegistry: { registerFilter, registerAnim } } = usesAnimation();
    registerFilter(activatableVars.filter);
    registerAnim(activatableVars.anim);
}
// .activated will be added after activating-animation done:
const selectorIfActivated = '.activated';
// [aria-checked],[aria-pressed],[aria-selected] = styled active, :checked = native active:
const selectorIfActivating = ':is([aria-checked]:not([aria-checked="false"]), [aria-pressed]:not([aria-pressed="false"]), [aria-selected]:not([aria-selected="false"]), :checked):not(.activated)';
// .passivating will be added after loosing active and will be removed after deactivating-animation done:
const selectorIfPassivating = '.passivating';
// if all above are not set => passivated:
const selectorIfPassivated = ':not(:is(.activated, [aria-checked]:not([aria-checked="false"]), [aria-pressed]:not([aria-pressed="false"]), [aria-selected]:not([aria-selected="false"]), :checked, .passivating))';
export const ifActivated = (styles) => rule(selectorIfActivated, styles);
export const ifActivating = (styles) => rule(selectorIfActivating, styles);
export const ifPassivating = (styles) => rule(selectorIfPassivating, styles);
export const ifPassivated = (styles) => rule(selectorIfPassivated, styles);
export const ifActive = (styles) => rule([selectorIfActivating, selectorIfActivated], styles);
export const ifPassive = (styles) => rule([selectorIfPassivating, selectorIfPassivated], styles);
export const ifActivePassivating = (styles) => rule([selectorIfActivating, selectorIfActivated, selectorIfPassivating], styles);
/**
 * Adds a capability of UI to be highlighted/selected/activated.
 * @param config  A configuration of `activatableRule`.
 * @returns A `ActivatableStuff` represents an activatable state.
 */
export const usesActivatable = (config) => {
    return {
        activatableRule: () => style({
            // configs:
            ...vars({
                /* supports for `usesPressAsActiveState()` */
                [activatableVars.filterActive]: config?.filterActive,
                [activatableVars.animActive]: config?.animActive,
                [activatableVars.animPassive]: config?.animPassive,
            }),
            // animation states:
            ...states([
                ifActivated({
                    ...vars({
                        [activatableVars.filter]: activatableVars.filterActive,
                    }),
                }),
                ifActivating({
                    ...vars({
                        [activatableVars.filter]: activatableVars.filterActive,
                        [activatableVars.anim]: activatableVars.animActive,
                    }),
                }),
                ifPassivating({
                    ...vars({
                        [activatableVars.filter]: activatableVars.filterActive,
                        [activatableVars.anim]: activatableVars.animPassive,
                    }),
                }),
            ]),
        }),
        activatableVars,
    };
};
export const markActive = (options) => {
    // options:
    const { outlined = false, mild = false } = options ?? {};
    return style({
        ...imports([
            setOutlined(outlined),
            setMild(mild), // kill mild     variant
        ]),
    });
};
const checkableCtrls = [
    'checkbox',
    'radio',
];
export const useActivatable = (props) => {
    // fn props:
    const propActive = usePropActive(props);
    const { tag, role } = useSemantic(props);
    // states:
    const [activated, setActivated] = useState(propActive); // true => active, false => passive
    const [animating, setAnimating] = useState(null); // null => no-animation, true => activating-animation, false => deactivating-animation
    /*
     * state is active/passive based on [controllable active]
     * [uncontrollable active] is not supported
     */
    const activeFn = propActive /*controllable*/;
    if (activated !== activeFn) { // change detected => apply the change & start animating
        const updateActivated = () => {
            setActivated(activeFn); // remember the last change
            setAnimating(activeFn); // start activating-animation/deactivating-animation
        };
        if (activeFn) {
            // update the state immediately:
            updateActivated();
        }
        else {
            // update the state a bit delayed:
            setTimeout(updateActivated, 2); // supports for toggle active => make sure the `.passivating` state runs *after* `.releasing` state
        } // if
    } // if
    // handlers:
    const handleAnimationEnd = useEvent((event) => {
        // conditions:
        if (event.target !== event.currentTarget)
            return; // ignores bubbling
        if (!/((?<![a-z])(active|passive)|(?<=[a-z])(Active|Passive))(?![a-z])/.test(event.animationName))
            return; // ignores animation other than (active|passive)[Foo] or boo(Active|Passive)[Foo]
        // clean up finished animation
        setAnimating(null); // stop activating-animation/deactivating-animation
    });
    return {
        active: activated,
        isVisible: activated || (animating !== null),
        class: (() => {
            // activating:
            if (animating === true)
                return null; // uses :checked or [aria-checked] or [aria-pressed] or [aria-selected]
            // passivating:
            if (animating === false)
                return 'passivating';
            // fully activated:
            if (activated)
                return 'activated';
            // fully passivated:
            return null;
        })(),
        props: (() => {
            if (!activated)
                return null;
            // use :checked if <input type="checkbox|radio">:
            if ((tag === 'input') && checkableCtrls.includes(props.type))
                return { checked: true };
            // use [aria-checked] if [role="checkbox|radio"]:
            if (role && checkableCtrls.includes(role))
                return { 'aria-checked': true };
            // use [aria-pressed] if <button> or [role="button"]:
            if ((tag === 'button') || (role === 'button'))
                return { 'aria-pressed': true };
            // else, use [aria-selected]:
            return { 'aria-selected': true };
        })(),
        handleAnimationEnd,
    };
};
export const useToggleActivatable = (props, changeEventTarget) => {
    // fn props:
    const { enabled, readOnly, active } = usePropAccessibility(props, undefined, undefined, null);
    // states:
    const [activeTg, setActiveTg] = useState(props.defaultActive ?? false);
    /*
     * state is active/passive based on [controllable active] (if set) and fallback to [uncontrollable active]
     */
    const activeFn = active /*controllable*/ ?? activeTg /*uncontrollable*/;
    // callbacks:
    const isDisabledOrReadOnly = (!enabled || readOnly);
    /*
          controllable : setActive(new) => update state(old => old) => trigger Event(new)
        uncontrollable : setActive(new) => update state(old => new) => trigger Event(new)
    */
    const triggerActiveChange = useEvent((active) => {
        Promise.resolve().then(() => {
            // fire change synthetic event:
            props.onActiveChange?.({ active });
            // fire change dom event:
            if (changeEventTarget?.current) {
                // *hack*: trigger `onChange` event:
                // side effect: toggles the [checked] prop:
                changeEventTarget.current.dispatchEvent(new PointerEvent('click', { bubbles: true, cancelable: true, composed: true }));
            } // if
        });
    });
    const setActive = useEvent((active) => {
        // conditions:
        if (isDisabledOrReadOnly)
            return; // control is disabled or readOnly => no response required
        const newActive = (typeof (active) === 'function') ? active(activeFn) : active;
        if (newActive === activeFn)
            return; // still the same => nothing to update
        // update:
        setActiveTg(newActive);
        triggerActiveChange(newActive);
    }); // a stable callback, the `setActive` guaranteed to never change
    const toggleActive = useEvent(() => {
        // conditions:
        if (isDisabledOrReadOnly)
            return; // control is disabled or readOnly => no response required
        const newActive = !activeFn;
        // update:
        setActiveTg(newActive);
        triggerActiveChange(newActive);
    }); // a stable callback, the `toggleActive` guaranteed to never change
    return [
        activeFn,
        setActive,
        toggleActive,
    ];
};
//#endregion activatable
